/*
 * Copyright 2011 NEHTA
 *
 * Licensed under the NEHTA Open Source (Apache) License; you may not use this
 * file except in compliance with the License. A copy of the License is in the
 * 'license.txt' file, which should be provided with this work.
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */
package au.gov.nehta.vendorlibrary.common;

import au.gov.nehta.common.utils.ArgumentUtils;
import org.w3c.dom.Document;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import javax.xml.namespace.QName;
import javax.xml.transform.stream.StreamSource;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * {@link au.gov.nehta.vendorlibrary.common.JAXBUtil} implementation for JAXB object marshal and unmarshal
 *
 */
public class JAXBUtil {

  /**
   * Constant for {@link java.util.logging.Logger} instance
   */
  private static final Logger LOG = Logger
    .getLogger(JAXBUtil.class.getName());

  /**
   * Marshall the provided object and return the XML content of the object.
   *
   * @param <T>
   * @param instanceClass
   *            The class instance of the object to be marshalled.
   * @param qName
   *            the namespace URI of the object to be marshalled. <br>
   *            <p>
   *            If the Namespace URI is null, it is set to
   *            XMLConstants.NULL_NS_URI. This value represents no explicitly
   *            defined Namespace as defined by the Namespaces in XML
   *            specification. This action preserves compatible behaviour with
   *            QName 1.0. Explicitly providing the XMLConstants.NULL_NS_URI
   *            value is the preferred coding style
   *            </P>
   * @param className
   *            the Class name of the object to be marshalled.
   * @param object
   *            the instance of the object. <br> {@link IllegalArgumentException}
   *            will be thrown if the object is null
   * @return XML source of the marshalled object.
   */
  @SuppressWarnings("unchecked")
  public static <T> String marshaller(Class instanceClass, String qName,
                                      String className, T object) {
    OutputStream os = null;
    if (object != null && instanceClass != null) {
      try {
        JAXBContext jc;
        Marshaller m;
        os = new ByteArrayOutputStream();
        jc = JAXBContext.newInstance(object.getClass());
        m = jc.createMarshaller();
        m.marshal(new JAXBElement(new QName(qName, className),
          instanceClass, object), os);
        return os.toString();
      } catch (JAXBException ex) {
        LOG.log(Level.SEVERE, null, ex);
      } finally {
        if (os != null) {
          try {
            os.flush();
            os.close();
          } catch (IOException ex) {
            LOG.log(Level.SEVERE, null, ex);
          }
        }
      }
    }
    return null;
  }

  /**
   * Marshalls the provided JAXWS object to String[XML String].
   *
   * @param <T>
   * @param instanceClass
   *            of type {@link Class}
   * @param className
   *            of the object to be marshalled
   * @param object
   *            instance value of T
   * @return Serialized value of T. Else returns null.
   */
  @SuppressWarnings("unchecked")
  public static <T> String marshaller(Class instanceClass, String className,
                                      T object) {
    return marshaller(instanceClass, null, className, object);
  }

  /**
   * Gets the JAXB/JAXWS instance for the provided XML data and jaxws class.
   *
   * @param <T>
   *            JAXWS generic type for unmarshalling xml data.
   * @param parameterObject
   *            of type {@link Class}
   * @param xmlSource
   *            of type {@link String}
   * @return <T> of type parameterObject
   */
  @SuppressWarnings("unchecked")
  public static <T> T unmarshaller(final Class parameterObject,
                                   final String xmlSource) {
    {
      InputStream is = null;
      if (parameterObject != null && xmlSource != null) {
        try {
          JAXBContext jc;
          Unmarshaller unM;
          try {
            is = new ByteArrayInputStream(xmlSource
              .getBytes("UTF-8"));
          } catch (UnsupportedEncodingException ex) {
            LOG
              .log(Level.SEVERE,
                "Unsupported encoding format", ex);
          }
          jc = JAXBContext.newInstance(parameterObject);
          unM = jc.createUnmarshaller();
          if (is != null) {
            JAXBElement<T> obj = unM.unmarshal(
              new StreamSource(is), parameterObject);
            return obj.getValue();
          }
        } catch (JAXBException ex) {
          LOG.log(Level.SEVERE, null, ex);
        } finally {
          try {
            if (is != null) {
              is.close();
            }
          } catch (IOException ex) {
            LOG.log(Level.SEVERE, null, ex);
          }
        }
      }
      return null;
    }
  }

  /**
   * Gets the JAXB/JAXWS instance for the provided XML data and jaxws class
   * object.
   *
   * @param <T>
   *            jaxws generic type for un-marshalling xml data.
   * @param parameterObject
   *            of type {@link Class}
   * @param doc
   *            // * @param xmlSource of type {@link String}
   * @return <T> of type parameterObject Note: throws
   *         IllegalArgumentExceptuion if either of the arguments are null.
   */
  @SuppressWarnings("unchecked")
  public static <T> T unmarshaller(final Class parameterObject,
                                   final Document doc) {
    // {
    ArgumentUtils.checkNotNull(parameterObject, "parameterObject");
    ArgumentUtils.checkNotNull(doc, "doc");
    doc.normalizeDocument();
    return (T) unmarshaller(parameterObject, DOMUtil.xmlToString(doc));
  }

}
